/*-
 * Copyright (C)2010..2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2010..2013 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
#endif /* not lint */
static const char rcsid[] = "$Id: blin.c,v 1.15 2014/04/09 14:43:11 babolo Exp $";

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <err.h>
#include "../BLINflag.h"

#define BLIN_MAININI   0x000001
#define BLIN_FLEGAL    BLIN_MASK
#define BLIN_MIMASK(A) (((u_int64_t)1 << A) - 1)
#define ERROUT(A,B)    do {ex = (A); errno = (B); goto out;} while(0)
#define BLIN_MAXTXTSTA 30
#define BLIN_MAXTXTDUM 88

blin_pamparam blin_pamparam_internal =
{ 0x00000000, 0
, { {NULL, BLIN_MODWARN | BLIN_MODEOAU, LOG_LOCAL7 | LOG_ERR}
  , {NULL, BLIN_MODWARN | BLIN_MODEOAU, LOG_LOCAL7 | LOG_WARNING}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
  , {NULL, BLIN_MODPRNT | BLIN_MODEOAU, LOG_LOCAL7 | LOG_DEBUG}
} };

__weak_reference(blin_pamparam_internal, blin_internal_flags);

static char *
fac[] =
{ "KERN"  , "USER"  , "MAIL"    , "DAEMON", "AUTH"  , "SYSLOG"  , "LPR"    , "NEWS"
, "UUCP"  , "CRON"  , "AUTHPRIV", "FTP"   , "NTP"   , "SECURITY", "CONSOLE", "reserved"
, "LOCAL0", "LOCAL1", "LOCAL2"  , "LOCAL3", "LOCAL4", "LOCAL5"  , "LOCAL6" , "LOCAL7"
};

static char *
pri[] = {"EMERG", "ALERT", "CRIT", "ERR", "WARNING", "NOTICE", "INFO", "DEBUG"};


static void
blin_init() {
    int i;

    if  (!(blin_pamparam_internal.flags & BLIN_MAININI)) {
        if  (blin_pamparam_internal.flags & BLIN_TRAC) {
            fprintf(stderr, "BLIN: +blin_init %08X\n", blin_pamparam_internal.flags);
        }
        for (i = 0; i < 8; i++) blin_pamparam_internal.w[i].file = stderr;
        blin_pamparam_internal.flags |= BLIN_MAININI;
        if  (blin_pamparam_internal.flags & BLIN_TRAC) {
            fprintf(stderr, "BLIN: -blin_init %08X\n", blin_pamparam_internal.flags);
}   }   }

static u_int32_t
blin_do(u_int32_t mode, u_int32_t w, u_int32_t legal, u_int32_t *f) {
    u_int32_t r = *f;

    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_do %08X %08X %08X %08X\n", mode, w, legal, *f);
    }
    if  (!(mode & BLIN_CTL_RISE)) {
        w = (mode & BLIN_CTL_NPRM) ? 0 : w;
    } else {
        w = (((((mode & BLIN_CTL_NPRM) ? r : w) << 1) & BLIN_MASK) | BLIN_BIT0);
    }
    if  (mode & BLIN_CTL_FNEG) w = ~w;
    w &= legal;
    switch (mode & (BLIN_CTL_NONE | BLIN_CTL_FAFM)) {
    case BLIN_CTL_FSET: *f |= w;                    break;
    case BLIN_CTL_FRST: *f &= ~w;                   break;
    case BLIN_CTL_FXOR: *f ^= w;                    break;
    case BLIN_CTL_FEQU: *f = (*f & ~legal) | w;     break;
    }
    if  (mode & BLIN_CTL_NPRE) r = *f;
    if  (mode & BLIN_CTL_RMSK) r &= legal;
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_do %08X\n", r);
    }
    return(r);
}

void *
blin_ctl(u_int32_t mode, ...) {
    void *res = NULL;
    u_int32_t m;
    va_list ap;
    int i;

    va_start(ap, mode);
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_ctl %08X\n", mode);
    }
    blin_init();
    switch (mode & BLIN_CTL_OPER) {
    case BLIN_CTL_FLAG: {
            u_int32_t w = va_arg(ap, u_int32_t);
            res = (void *)(size_t)blin_do(mode, w, BLIN_FLEGAL, &blin_pamparam_internal.flags);
        }
        break;
    case BLIN_CTL_WHAT: {
            u_int32_t w = va_arg(ap, u_int32_t);
            for (m = BLIN_BIT0, i = 0; m; m <<= 1, ++i) {
                if  (mode & m) {
                    res = (void *)(size_t)blin_do(mode, w, ~0, &blin_pamparam_internal.w[i].what);
        }   }   }
        break;
    case BLIN_CTL_FNAM: {
            char *f = va_arg(ap, char *);
            FILE *w = fopen(f, "a");
            for (m = BLIN_BIT0, i = 0; m < BLIN_MASK; m <<= 1, ++i) {
                if  (mode & m) {
                    res = blin_pamparam_internal.w[i].file;
                    if  (!(mode & BLIN_CTL_NONE)) blin_pamparam_internal.w[i].file = w;
        }   }   }
        break;
    case BLIN_CTL_FDES: {
            int f = va_arg(ap, int);
            FILE *w = fdopen(f, "a");
            for (m = BLIN_BIT0, i = 0; m < BLIN_MASK; m <<= 1, ++i) {
                if  (mode & m) {
                    res = blin_pamparam_internal.w[i].file;
                    if  (!(mode & BLIN_CTL_NONE)) blin_pamparam_internal.w[i].file = w;
        }   }   }
        break;
    case BLIN_CTL_FILE: {
            void *w = va_arg(ap, void *);
            for (m = BLIN_BIT0, i = 0; m < BLIN_MASK; m <<= 1, ++i) {
                if  (mode & m) {
                    res = blin_pamparam_internal.w[i].file;
                    if  (!(mode & BLIN_CTL_NONE)) blin_pamparam_internal.w[i].file = w;
        }   }   }
        break;
    case BLIN_CTL_PRIO: {
            u_int32_t w = va_arg(ap, u_int32_t);
            for (m = BLIN_BIT0, i = 0; m < BLIN_MASK; m <<= 1, ++i) {
                if  (mode & m) {
                    res = (void *)BLIN_I(blin_pamparam_internal.w[i].prio);
                    if  (!(mode & BLIN_CTL_NONE)) blin_pamparam_internal.w[i].prio = w;
        }   }   }
        break;
    case BLIN_CTL_RCSI: {
            res = (void *)rcsid;
        }
        break;
    case BLIN_CTL_VERS: {
            res = (void *)VERS;
        }
        break;
    case BLIN_CTL_DUMP: {
            int k = 0;
        ;   m = blin_pamparam_internal.flags;
        ;   fprintf(stderr, "blin_pamparam_internal %08X=<", m);
        ;   if  (m & ~(BLIN_MASK | BLIN_MAININI)) {
        ;       fprintf(stderr, "%sunkn", k++ ? " | " : "");
        ;   }
        ;   if  (m & BLIN_MAININI) {
        ;       fprintf(stderr, "%sMAININI", k++ ? " | " : "");
        ;   }
        ;   for (i = 0; m & BLIN_BIT0; ++i) {
        ;       m >>= 1;
        ;       if  (!(m & BLIN_BIT0)) {
        ;           fprintf(stderr, "%sGEN%u", k++ ? " | " : "", i);
        ;           break;
        ;   }   }
        ;   for (++i; m & BLIN_MASK; m >>= 1, ++i) {
        ;       if  (m & BLIN_BIT0) {
        ;           fprintf(stderr, "%sBIT%u", k++ ? " | " : "", i);
        ;   }   }
        ;   if  (blin_pamparam_internal.dummy) {
        ;       fprintf(stderr, "> !!! %08X !!!\n", blin_pamparam_internal.dummy);
        ;   } else {
        ;       fprintf(stderr, ">\n");
        ;   }
        ;   for (i = 0; i < 8; ++i) {
                k = 0;
        ;       m = blin_pamparam_internal.w[i].what;
        ;       fprintf(stderr, "  %u %08X=<", i, m);
        ;       if  (m & ~( BLIN_MODWARN
                          | BLIN_MODPRNT
                          | BLIN_MODLOGR
                          | BLIN_MODEOAU
                          | BLIN_MODEOCR
                          | BLIN_MODEOLF
                          | BLIN_MODTIME
                    )     ) {
        ;           fprintf(stderr, "%sunkn", k++ ? " | " : "");
        ;       }
        ;       if  (m & BLIN_MODWARN) fprintf(stderr, "%sMODWARN", k++ ? " | " : "");
        ;       if  (m & BLIN_MODPRNT) fprintf(stderr, "%sMODPRNT", k++ ? " | " : "");
        ;       if  (m & BLIN_MODLOGR) fprintf(stderr, "%sMODLOGR", k++ ? " | " : "");
        ;       if  (m & BLIN_MODEOAU) fprintf(stderr, "%sMODEOAU", k++ ? " | " : "");
        ;       if  (m & BLIN_MODEOCR) fprintf(stderr, "%sMODEOCR", k++ ? " | " : "");
        ;       if  (m & BLIN_MODEOLF) fprintf(stderr, "%sMODEOLF", k++ ? " | " : "");
        ;       if  (m & BLIN_MODTIME) fprintf(stderr, "%sMODTIME", k++ ? " | " : "");
        ;       k = 0;
        ;       m = blin_pamparam_internal.w[i].prio;
        ;       fprintf(stderr, "> %08X=<", m);
        ;       if  (m & ~(LOG_PRIMASK | LOG_FACMASK)) {
        ;           fprintf(stderr, "%sunkn", k++ ? " | " : "");
        ;       }
        ;       if  (((m & LOG_FACMASK) >> 3) >= LOG_NFACILITIES) {
        ;           fprintf(stderr, "%sbig", k++ ? " | " : "");
        ;       }
        ;       fprintf(stderr, "%s%s", k++ ? " | " : "", fac[(m & LOG_FACMASK) >> 3]);
        ;       fprintf(stderr, "%s%s", k++ ? " | " : "", pri[m & LOG_PRIMASK]);
        ;       fprintf(stderr, ">");
        ;       if  (blin_pamparam_internal.w[i].file == stdin) {
        ;           fprintf(stderr, "  file=stdin\n");
        ;       } else if (blin_pamparam_internal.w[i].file == stdout) {
        ;           fprintf(stderr, "  file=stdout\n");
        ;       } else if (blin_pamparam_internal.w[i].file == stderr) {
        ;           fprintf(stderr, "  file=stderr\n");
        ;       } else {
        ;           fprintf(stderr, "  file=%"BLIN_X"\n", BLIN_I(blin_pamparam_internal.w[i].file));
    }   }   }   }
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_ctl %08X\n", (u_int32_t)res);
    }
    va_end(ap);
    return(res);
}

static void
doout(int level, char *q) {
    u_int32_t ctl = blin_pamparam_internal.w[level].what;

    if  (ctl & BLIN_MODWARN) {
        warnx("%s", q);
    }
    if  (ctl & BLIN_MODPRNT) {
        fprintf(blin_pamparam_internal.w[level].file, "%s\n", q);
    }
    if  (ctl & BLIN_MODLOGR) {
        syslog(blin_pamparam_internal.w[level].prio, "%s", q);
}   }

static void
dotime(int level) {
    u_int32_t ctl = blin_pamparam_internal.w[level].what;
    struct timespec tp;

    if  (ctl & BLIN_MODTIME) {
        if  (0 > clock_gettime(CLOCK_REALTIME, &tp)) {
            ctl &= ~BLIN_MODTIME;
        } else {
            if  (ctl & BLIN_MODWARN) {
                warnx("%10"BLIN_D".%09ld", tp.tv_sec, tp.tv_nsec);
            }
            if  (ctl & BLIN_MODPRNT) {
                fprintf( blin_pamparam_internal.w[level].file, "%10"BLIN_D".%09ld\n"
                       , tp.tv_sec, tp.tv_nsec
                       );
            }
            if  (ctl & BLIN_MODLOGR) {
                syslog( blin_pamparam_internal.w[level].prio, "%10"BLIN_D".%09ld"
                      , tp.tv_sec, tp.tv_nsec
                      );
}   }   }   }

int
blin_cry1(int ctl, const char *format, ...) {
    struct timespec tp;
    u_int32_t control;
    size_t l = 0;
    va_list ap;

    va_start(ap, format);
    blin_init();
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_cry1 %08X\n", ctl);
    }
    control = blin_pamparam_internal.w[ctl & 7].what;
    if  (control & BLIN_MODTIME) {
        if  (0 > clock_gettime(CLOCK_REALTIME, &tp)) control &= ~BLIN_MODTIME;
    }
    if  (control & BLIN_MODWARN) {
        if  (control & BLIN_MODTIME) {
            warnx("%10"BLIN_D".%09ld ", tp.tv_sec, tp.tv_nsec);
        }
        if  (ctl & 8) vwarn(format, ap); else vwarnx(format, ap);
    }
    if  (control & BLIN_MODPRNT) {
        if  (control & BLIN_MODTIME) {
            fprintf( blin_pamparam_internal.w[ctl & 7].file, "%10"BLIN_D".%09ld "
                   , tp.tv_sec, tp.tv_nsec
                   )
            ;
        }
        vfprintf(blin_pamparam_internal.w[ctl & 7].file, format, ap);
        if  (ctl & 8) {
            fprintf( blin_pamparam_internal.w[ctl & 7].file
                   , ": %s"
                   , sys_errlist[(errno > sys_nerr) ? 0 : errno]
                   );
            control |= BLIN_MODEOLF;
    }   }
    if  (control & BLIN_MODLOGR) {
        if  (control & BLIN_MODTIME) {
            syslog( blin_pamparam_internal.w[ctl & 7].prio, "%10"BLIN_D".%09ld "
                   , tp.tv_sec, tp.tv_nsec
                   )
            ;
        }
        vsyslog(blin_pamparam_internal.w[ctl & 7].prio, format, ap);
        if  (ctl & 8) syslog(blin_pamparam_internal.w[ctl & 7].prio, ": %m");
    }
    if  (control & BLIN_MODEOAU) {
        if  (format) l = strlen(format);
        if  (l && (format[l - 1] != '\n')) control |= BLIN_MODEOLF;
    }
    if  (!(~control & (BLIN_MODEOCR | BLIN_MODPRNT))) {
        fprintf(blin_pamparam_internal.w[ctl & 7].file, "\r");
    }
    if  (!(~control & (BLIN_MODEOLF | BLIN_MODPRNT))) {
        fprintf(blin_pamparam_internal.w[ctl & 7].file, "\n");
    }
    va_end(ap);
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_cry1\n");
    }
    return(0);
}

int
blin_stateheader(blin_statectl staflag, const char actnames[][3]) {
    int i, ex = EX_OK;
    u_int32_t ctl;
    ssize_t l, t;
    char *q, *p;

    blin_init();
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_stateheader %08X\n", staflag.statectl);
    }
    dotime(staflag.level);
    ctl = blin_pamparam_internal.w[staflag.level].what;
    l = strlen("control-");
    if  (!staflag.allign) {
        for (i = 0; i < 24; ++i) if (actnames[i][0] != '-') l += 3;
    } else {
        l += 3 * staflag.allign;
    }
    l += strlen("nst rs cl st --rg32-- --offs--");
        /*     "---"   */
    if  (!(q = calloc(1, l + 3))) {
        ifBLIN_QX0("malloc");
        ERROUT(EX_OSERR, ENOMEM);
    }
    t = l;
    strncpy(q, "control-", t);
        /*  q  '\0'  strncpy()   calloc() */
    p = index(q, 0);
    if  (0 >= (t = q + l - p)) {
        ifBLIN_QX0("strcat #0");
        ERROUT(EX_SOFTWARE, EDOOFUS);
    }
    if  (!staflag.allign) {
        for (i = 0; i < 24; ++i) {
            if  (actnames[i][0] != '-') {
                *(p++) = ' ';
                *(p++) = actnames[i][0];
                *(p++) = actnames[i][1];
                if  (0 >= (t -= 3)) {
                    ifBLIN_QX0("strcat #1 %d", i);
                    ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }   }
    } else  {
        for (i = 0; i < staflag.allign; ++i) {
            *(p++) = ' ';
            *(p++) = '-';
            *(p++) = '-';
            if  (0 >= (t -= 3)) {
                ifBLIN_QX0("strcat #2 %d", i);
                ERROUT(EX_SOFTWARE, EDOOFUS);
    }   }   }
        /*    (++t)      */
    strncat(q, "nst rs cl st --rg32-- --offs--", ++t);
    p = index(q, 0);
    if  (0 > (t = q + l - p)) {
        ifBLIN_QX0("strcat #3 real %"BLIN_D" > %"BLIN_D" alloc", BLIN_I(p - q), BLIN_I(l));
        ERROUT(EX_SOFTWARE, EDOOFUS);
    } else if (0 < (t = q + l - p)) {
        ifBLIN_QX1("strcat #3 real %"BLIN_D" < %"BLIN_D" alloc", BLIN_I(p - q), BLIN_I(l));
    }
    doout(staflag.level, q);
out:
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_stateheader %d\n", ex);
    }
    return(ex);
}

int
blin_statebody( blin_statectl  staflag
              , const char     actnames[][3]   /*                               */
              , const char     statenames[][3] /*                              */
              , const char     clasnname[3]    /*                            */
              , const char    *in           /*                           */
              , size_t         offset       /*                    */
              , u_int64_t      control      /*                               */
              , u_int32_t      state        /*                                      */
              , u_int32_t      reg          /*   32                     */
              , u_char         byte         /*   8                      */
              ) {
    u_int64_t C_state = BLIN_MIMASK(staflag.staten);
    u_int64_t flymask, contrana;
    int i, k, ex = EX_OK;
    u_int32_t ctl;
    ssize_t l, t;
    char *q, *p;

    blin_init();
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_statebody %08X\n", staflag.statectl);
    }
    if  (staflag.revert) {
        flymask = (u_int64_t)1 << (staflag.staten + staflag.actinn - 1);
    } else {
        flymask = (u_int64_t)1 << (staflag.staten);
    }
    ctl = blin_pamparam_internal.w[staflag.level].what;
    l = 9;
    if  (!staflag.allign) {
        for (i = 0; i < 24; ++i) if (actnames[i][0] != '-') l += 3;
    } else {
        l += 3 * staflag.allign;
    }
    l += 3 + 3 + 3 + 3 + 9;
#ifdef __amd64__
    l += 13;
#else
    l += 9;
#endif
    l += BLIN_MAXTXTSTA + 2;
    if  (!(q = calloc(1, l + 3))) {
        ifBLIN_QX0("malloc");
        ERROUT(EX_OSERR, ENOMEM);
    }
    p = q + snprintf(q, l, "%08X ", (u_int32_t)control);
    if  (0 >= (t = q + l - p)) {
        ifBLIN_QX0("strcat #0");
        ERROUT(EX_SOFTWARE, EDOOFUS);
    }
    if  (!staflag.allign) {
        for (i = 0, contrana = control, k = 0; i < staflag.actinn; ++i) {
            if  (actnames[i][0] == '-') {
                if  (contrana & flymask) {
                    ++k;
                }
            } else {
                if  (contrana & flymask) {
                    *(p++) = actnames[i][0];
                    *(p++) = actnames[i][1];
                } else {
                    *(p++) = ' ';
                    *(p++) = ' ';
                }
                if  (k && (i == (staflag.actinn - 1))) {
                    *(p++) = '-';
                } else {
                    *(p++) = ' ';
                }
                if  (0 >= (t -= 3)) {
                    ifBLIN_QX0("strcat #1 %d", i);
                    ERROUT(EX_SOFTWARE, EDOOFUS);
            }   }
            if  (staflag.revert) contrana <<= 1; else contrana >>= 1;
        }
    } else {
        for (i = 0, contrana = control, k = 0; i < staflag.actinn; ++i) {
            if  (contrana & flymask) {
                *(p++) = actnames[i][0];
                *(p++) = actnames[i][1];
                *(p++) = ' ';
                if  (0 >= (t -= 3)) {
                    ifBLIN_QX0("strcat #1 %d", i);
                    ERROUT(EX_SOFTWARE, EDOOFUS);
                }
                k++;
            }
            if  (staflag.revert) contrana <<= 1; else contrana >>= 1;
        }
        for (; k < staflag.allign; ++k) {
            *(p++) = ' ';
            *(p++) = ' ';
            *(p++) = ' ';
            if  (0 >= (t -= 3)) {
                ifBLIN_QX0("strcat #2 %d", i);
                ERROUT(EX_SOFTWARE, EDOOFUS);
    }   }   }
    p += snprintf( p, t, "%s %02X %s %s %08X %"BLIN_X" ~"
                 , statenames[control & C_state]
                 , byte & 0x00FF
                 , clasnname
                 , statenames[state]
                 , reg
                 , BLIN_I(offset)
                 );
    if  (0 >= (t = q + l - p)) {
        ifBLIN_QX0("strcat #3");
        ERROUT(EX_SOFTWARE, EDOOFUS);
    }
    for (i = 0; (i <= BLIN_MAXTXTSTA) && in[i]; ++i) {
        if  ((staflag.gtequ) ? (in[i] < staflag.bordsy) : (in[i] == staflag.bordsy)) break;
        *(p++) = in[i];
        --t;
        if  (0 > t) {
            ifBLIN_QX0("strcat #4");
            ERROUT(EX_SOFTWARE, EDOOFUS);
    }   }
    if  (0 < t) {
        *(p++) = '~';
        --t;
    } else if (0 > t) {
        ifBLIN_QX0("strcat #5");
        ERROUT(EX_SOFTWARE, EDOOFUS);
    }
    doout(staflag.level, q);
out:
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_statebody %d\n", ex);
    }
    return(ex);
}

static char
z(u_char u) {
    char s;

    if  (u < ' ') s = '.';
    else if ((u < 0x7F) || (u >= (u_char)'') || (u == (u_char)'') || (u == (u_char)'')) s = u;
    else s = '.';
    return(s);
}

int
blin_dumb(int mode, void *buf, ssize_t len) {
    char q[BLIN_MAXTXTDUM + 3], *p;
    ssize_t t;
    int i, ex = EX_OK;
    u_int32_t control;
    u_char *c;

    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_dumb %08X %"BLIN_D"\n", mode, len);
    }
    c = buf;
    dotime(mode & 7);
    control = blin_pamparam_internal.w[mode & 7].what;
    while (len > 0) {
        p = q + snprintf(q, BLIN_MAXTXTDUM, "%"BLIN_X, BLIN_I(c));
        if  (0 >= (t = q + BLIN_MAXTXTDUM - p)) {
            ifBLIN_QX0("strcat #0");
            ERROUT(EX_SOFTWARE, EDOOFUS);
        }
        for (i = 0; i < 16; ++i) {
            if  (i && !(i & 3)) {
                *(p++) = ' ';
                t--;
                if  (0 > t) {
                    ifBLIN_QX0("strcat #1 %d", i);
                    ERROUT(EX_SOFTWARE, EDOOFUS);
            }   }
            if  (i < len) {
                p += snprintf(p, t, " %02X", c[i]);
            } else if (mode & BLIN_DU_TEXT) {
                *(p++) = ' ';
                *(p++) = ' ';
                *(p++) = ' ';
            }
            if  (0 >= (t = q + BLIN_MAXTXTDUM - p)) {
                ifBLIN_QX0("strcat #2 %d", i);
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        if  (mode & BLIN_DU_TEXT) {
            *(p++) = ' ';
            *(p++) = '|';
            if  (0 >= (t -= 2)) {
                ifBLIN_QX0("strcat #3");
                ERROUT(EX_SOFTWARE, EDOOFUS);
            }
            for (i = 0; i < 16; ++i) {
                if  (i < len) {
                    *(p++) = z(c[i]);
                    if  (0 >= (--t)) {
                        ifBLIN_QX0("strcat #4");
                        ERROUT(EX_SOFTWARE, EDOOFUS);
            }   }   }
            *(p++) = '|';
            if  (0 >= (--t)) {
                ifBLIN_QX0("strcat #5");
                ERROUT(EX_SOFTWARE, EDOOFUS);
        }   }
        *(p++) = '\0';
        --t;
        doout(mode & 7, q);
        if  (len >= 16) {
            c += 16;
            len -= 16;
        } else {
            c += len;
            len = 0;
    }   }
out:
    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: -blin_dumb %d\n", ex);
    }
    return(ex);
}

#if 0
void
blin_dumb(int mode, void *buf, size_t len) {
    u_int32_t control;
    char format[128];
    u_char *c;
    size_t p;

    if  (blin_pamparam_internal.flags & BLIN_TRAC) {
        fprintf(stderr, "BLIN: +blin_dumb %08X %"BLIN_D"\n", mode, len);
    }
    c = buf;
    dotime(mode & 7);
    control = blin_pamparam_internal.w[mode & 7].what;
    p = XFORM((mode & BLIN_DU_TEXT) ? 17 : 16);
    memcpy(format, fullformat, p);
    format[p++] = '\n';
    format[p++] = '\0';
    if  (control & BLIN_MODWARN) {
    ;   while (len >= 16) {
    ;       if  (mode & BLIN_DU_TEXT) {
    ;           warnx( format, BLIN_I(c)
                     , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                     , c[8], c[9], c[10], c[11], c[12], c[13], c[14], c[15]
                     , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                     , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12]), z(c[13]), z(c[14]), z(c[15])
                     );
    ;       } else {
    ;           warnx( format, BLIN_I(c)
                     , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                     , c[8], c[9], c[10], c[11], c[12], c[13], c[14], c[15]
                     );
    ;       }
    ;       len -= 16;
    ;       c += 16;
    ;   }
    ;   if  (len) {
    ;       coformat(mode, len, format);
    ;       if  (mode & BLIN_DU_TEXT) {
    ;       ;   switch(len) {
                case 15: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10], c[11], c[12], c[13], c[14]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                              , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12]), z(c[13]), z(c[14])
                              );
    ;       ;            break;
                case 14: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10], c[11], c[12], c[13]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                              , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12]), z(c[13])
                              );
    ;       ;            break;
                case 13: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10], c[11], c[12]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                              , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12])
                              );
    ;       ;            break;
                case 12: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10], c[11]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                              , z(c[8]), z(c[9]), z(c[10]), z(c[11])
                              );
    ;       ;            break;
                case 11: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                              , z(c[8]), z(c[9]), z(c[10])
                              );
    ;       ;            break;
                case 10: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                              , z(c[8]), z(c[9])
                              );
    ;       ;            break;
                case 9:  warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                              , z(c[8])
                              );
    ;       ;            break;
                case 8:  warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                              );
    ;       ;            break;
                case 7:  warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                              );
    ;       ;            break;
                case 6:  warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5])
                              );
    ;       ;            break;
                case 5:  warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4])
                              );
    ;       ;            break;
                case 4:  warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3]
                              , z(c[0]), z(c[1]), z(c[2]), z(c[3])
                              );
    ;       ;            break;
                case 3:  warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2]
                              , z(c[0]), z(c[1]), z(c[2])
                              );
    ;       ;            break;
                case 2:  warnx(format, BLIN_I(c), c[0], c[1], z(c[0]), z(c[1]));
    ;       ;            break;
                case 1:  warnx(format, BLIN_I(c), c[0], z(c[0]));
    ;       ;            break;
    ;       ;   }
    ;       } else {
    ;       ;   switch(len) {
                case 15: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10], c[11], c[12], c[13], c[14]
                              );
    ;       ;            break;
                case 14: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10], c[11], c[12], c[13]
                              );
    ;       ;            break;
                case 13: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10], c[11], c[12]
                              );
    ;       ;            break;
                case 12: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10], c[11]
                              );
    ;       ;            break;
                case 11: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9], c[10]
                              );
    ;       ;            break;
                case 10: warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8], c[9]
                              );
    ;       ;            break;
                case 9:  warnx( format, BLIN_I(c)
                              , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                              , c[8]
                              );
    ;       ;            break;
                case 8:  warnx(format, BLIN_I(c), c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]);
    ;       ;            break;
                case 7:  warnx(format, BLIN_I(c), c[0], c[1], c[2], c[3], c[4], c[5], c[6]);
    ;       ;            break;
                case 6:  warnx(format, BLIN_I(c), c[0], c[1], c[2], c[3], c[4], c[5]);
    ;       ;            break;
                case 5:  warnx(format, BLIN_I(c), c[0], c[1], c[2], c[3], c[4]);
    ;       ;            break;
                case 4:  warnx(format, BLIN_I(c), c[0], c[1], c[2], c[3]);
    ;       ;            break;
                case 3:  warnx(format, BLIN_I(c), c[0], c[1], c[2]);
    ;       ;            break;
                case 2:  warnx(format, BLIN_I(c), c[0], c[1]);
    ;       ;            break;
                case 1:  warnx(format, BLIN_I(c), c[0]);
    ;       ;            break;
    }   }   }   }
    if  (control & BLIN_MODPRNT) {
    ;   while (len >= 16) {
    ;       if  (mode & BLIN_DU_TEXT) {
    ;           fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                       , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                       , c[8], c[9], c[10], c[11], c[12], c[13], c[14], c[15]
                       , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                       , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12]), z(c[13]), z(c[14]), z(c[15])
                       );
    ;       } else {
    ;           fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                       , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                       , c[8], c[9], c[10], c[11], c[12], c[13], c[14], c[15]
                       );
    ;       }
    ;       len -= 16;
    ;       c += 16;
    ;   }
    ;   if  (len) {
    ;       coformat(mode, len, format);
    ;       if  (mode & BLIN_DU_TEXT) {
    ;       ;   switch(len) {
                case 15: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10], c[11], c[12], c[13], c[14]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3])
                                , z(c[4]), z(c[5]), z(c[6]), z(c[7])
                                , z(c[8]), z(c[9]), z(c[10]), z(c[11])
                                , z(c[12]), z(c[13]), z(c[14])
                                );
    ;       ;            break;
                case 14: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10], c[11], c[12], c[13]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                                , z(c[7]), z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12]), z(c[13])
                                );
    ;       ;            break;
                case 13: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10], c[11], c[12]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                                , z(c[7]), z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12])
                                );
    ;       ;            break;
                case 12: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10], c[11]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                                , z(c[7]), z(c[8]), z(c[9]), z(c[10]), z(c[11])
                                );
    ;       ;            break;
                case 11: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                                , z(c[7]), z(c[8]), z(c[9]), z(c[10])
                                );
    ;       ;            break;
                case 10: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                                , z(c[7]), z(c[8]), z(c[9])
                                );
    ;       ;            break;
                case 9:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                                , z(c[7]), z(c[8])
                                );
    ;       ;            break;
                case 8:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                                , z(c[7])
                                );
    ;       ;            break;
                case 7:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                                );
    ;       ;            break;
                case 6:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5])
                                );
    ;       ;            break;
                case 5:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4])
                                );
    ;       ;            break;
                case 4:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3]
                                , z(c[0]), z(c[1]), z(c[2]), z(c[3])
                                );
    ;       ;            break;
                case 3:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2]
                                , z(c[0]), z(c[1]), z(c[2])
                                );
    ;       ;            break;
                case 2:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], z(c[0]), z(c[1])
                                );
    ;       ;            break;
                case 1:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], z(c[0])
                                );
    ;       ;            break;
    ;       ;   }
    ;       } else {
    ;       ;   switch(len) {
                case 15: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10], c[11], c[12], c[13], c[14]
                                );
    ;       ;            break;
                case 14: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10], c[11], c[12], c[13]
                                );
    ;       ;            break;
                case 13: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10], c[11], c[12]
                                );
    ;       ;            break;
                case 12: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10], c[11]
                                );
    ;       ;            break;
                case 11: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9], c[10]
                                );
    ;       ;            break;
                case 10: fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8], c[9]
                                );
    ;       ;            break;
                case 9:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                , c[8]
                                );
    ;       ;            break;
                case 8:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                                );
    ;       ;            break;
                case 7:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5], c[6]
                                );
    ;       ;            break;
                case 6:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4], c[5]
                                );
    ;       ;            break;
                case 5:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3], c[4]
                                );
    ;       ;            break;
                case 4:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2], c[3]
                                );
    ;       ;            break;
                case 3:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1], c[2]
                                );
    ;       ;            break;
                case 2:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0], c[1]
                                );
    ;       ;            break;
                case 1:  fprintf( blin_pamparam_internal.w[mode & 7].file, format, BLIN_I(c)
                                , c[0]
                                );
    ;       ;            break;
    }   }   }   }
    if  (control & BLIN_MODLOGR) {
    ;   while (len >= 16) {
    ;       if  (mode & BLIN_DU_TEXT) {
    ;           syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                      , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                      , c[8], c[9], c[10], c[11], c[12], c[13], c[14], c[15]
                      , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                      , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12]), z(c[13]), z(c[14]), z(c[15])
                      );
    ;       } else {
    ;           syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                      , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                      , c[8], c[9], c[10], c[11], c[12], c[13], c[14], c[15]
                      );
    ;       }
    ;       len -= 16;
    ;       c += 16;
    ;   }
    ;   if  (len) {
    ;       coformat(mode, len, format);
    ;       if  (mode & BLIN_DU_TEXT) {
    ;       ;   switch(len) {
                case 15: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10], c[11], c[12], c[13], c[14]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                               , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12]), z(c[13]), z(c[14])
                               );
    ;       ;            break;
                case 14: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10], c[11], c[12], c[13]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                               , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12]), z(c[13])
                               );
    ;       ;            break;
                case 13: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10], c[11], c[12]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                               , z(c[8]), z(c[9]), z(c[10]), z(c[11]), z(c[12])
                               );
    ;       ;            break;
                case 12: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10], c[11]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                               , z(c[8]), z(c[9]), z(c[10]), z(c[11])
                               );
    ;       ;            break;
                case 11: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                               , z(c[8]), z(c[9]), z(c[10])
                               );
    ;       ;            break;
                case 10: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                               , z(c[8]), z(c[9])
                               );
    ;       ;            break;
                case 9:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                               , z(c[8])
                               );
    ;       ;            break;
                case 8:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6]), z(c[7])
                               );
    ;       ;            break;
                case 7:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5]), z(c[6])
                               );
    ;       ;            break;
                case 6:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4]), z(c[5])
                               );
    ;       ;            break;
                case 5:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4]
                               , z(c[0]), z(c[1]), z(c[2]), z(c[3]), z(c[4])
                               );
    ;       ;            break;
                case 4:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], z(c[0]), z(c[1]), z(c[2]), z(c[3])
                               );
    ;       ;            break;
                case 3:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], z(c[0]), z(c[1]), z(c[2])
                               );
    ;       ;            break;
                case 2:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], z(c[0]), z(c[1])
                               );
    ;       ;            break;
                case 1:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], z(c[0])
                               );
    ;       ;            break;
    ;       ;   }
    ;       } else {
    ;       ;   switch(len) {
                case 15: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10], c[11], c[12], c[13], c[14]
                               );
    ;       ;            break;
                case 14: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10], c[11], c[12], c[13]
                               );
    ;       ;            break;
                case 13: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10], c[11], c[12]
                               );
    ;       ;            break;
                case 12: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10], c[11]
                               );
    ;       ;            break;
                case 11: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9], c[10]
                               );
    ;       ;            break;
                case 10: syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8], c[9]
                               );
    ;       ;            break;
                case 9:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               , c[8]
                               );
    ;       ;            break;
                case 8:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6], c[7]
                               );
    ;       ;            break;
                case 7:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5], c[6]
                               );
    ;       ;            break;
                case 6:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4], c[5]
                               );
    ;       ;            break;
                case 5:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3], c[4]
                               );
    ;       ;            break;
                case 4:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2], c[3]
                               );
    ;       ;            break;
                case 3:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1], c[2]
                               );
    ;       ;            break;
                case 2:  syslog( blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c)
                               , c[0], c[1]
                               );
    ;       ;            break;
                case 1:  syslog(blin_pamparam_internal.w[mode & 7].prio, format, BLIN_I(c), c[0]);
    ;       ;            break;
}   }   }   }   }
#endif
